@@ -27,6 +27,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
+#include <errno.h>
 
 #include <glib.h>
 
@@ -43,10 +44,10 @@ struct stk_data {
 	GAtChat *chat;
 };
 
-static const char *stke_prefix[] = { "%STKE:", NULL };
 static const char *none_prefix[] = { NULL };
+static const char *stke_prefix[] = { "*STKE:", NULL };
 
-static void mbm_stke_cb(gboolean ok, GAtResult *result, gpointer user_data)
+static void stke_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cb_data *cbd = user_data;
 	ofono_stk_envelope_cb_t cb = cbd->cb;
@@ -55,24 +56,25 @@ static void mbm_stke_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	const guint8 *pdu = NULL;
 	gint len = 0;
 
+	DBG("");
+
 	decode_at_error(&error, g_at_result_final_response(result));
 
-	if (!ok) {
-		cb(&error, NULL, 0, cbd->data);
-		return;
-	}
+	if (!ok)
+		goto done;
 
 	g_at_result_iter_init(&iter, result);
 
-	if (g_at_result_iter_next(&iter, "*STKE:") == TRUE)
-		if (g_at_result_iter_next_hexstring(&iter, &pdu, &len) == FALSE)
-			goto error;
+	if (g_at_result_iter_next(&iter, "*STKE:") == FALSE)
+		goto done;
 
+	/* Response data is optional */
+	g_at_result_iter_next_hexstring(&iter, &pdu, &len);
+
+	DBG("len %d", len);
+
+done:
 	cb(&error, pdu, len, cbd->data);
-	return;
-
-error:
-	CALLBACK_WITH_FAILURE(cb, NULL, 0, cbd->data);
 }
 
 static void mbm_stk_envelope(struct ofono_stk *stk, int length,
@@ -82,7 +84,9 @@ static void mbm_stk_envelope(struct ofono_stk *stk, int length,
 	struct stk_data *sd = ofono_stk_get_data(stk);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char *buf = g_try_new(char, 64 + length * 2);
-	int len, ret;
+	int len;
+
+	DBG("");
 
 	if (!cbd || !buf)
 		goto error;
@@ -92,14 +96,13 @@ static void mbm_stk_envelope(struct ofono_stk *stk, int length,
 		len += sprintf(buf + len, "%02hhX", *command++);
 	len += sprintf(buf + len, "\"");
 
-	ret = g_at_chat_send(sd->chat, buf, stke_prefix,
-				mbm_stke_cb, cbd, g_free);
+	DBG("%s", buf);
 
-	g_free(buf);
-	buf = NULL;
-
-	if (ret > 0)
+	if (g_at_chat_send(sd->chat, buf, stke_prefix,
+					stke_cb, cbd, g_free) > 0) {
+		g_free(buf);
 		return;
+	}
 
 error:
 	g_free(buf);
@@ -108,12 +111,14 @@ error:
 	CALLBACK_WITH_FAILURE(cb, NULL, 0, data);
 }
 
-static void mbm_stkr_cb(gboolean ok, GAtResult *result, gpointer user_data)
+static void stkr_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cb_data *cbd = user_data;
 	ofono_stk_generic_cb_t cb = cbd->cb;
 	struct ofono_error error;
 
+	DBG("");
+
 	decode_at_error(&error, g_at_result_final_response(result));
 	cb(&error, cbd->data);
 }
@@ -125,7 +130,9 @@ static void mbm_stk_terminal_response(struct ofono_stk *stk, int length,
 	struct stk_data *sd = ofono_stk_get_data(stk);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char *buf = g_try_new(char, 64 + length * 2);
-	int len, ret;
+	int len;
+
+	DBG("");
 
 	if (!cbd || !buf)
 		goto error;
@@ -135,14 +142,13 @@ static void mbm_stk_terminal_response(struct ofono_stk *stk, int length,
 		len += sprintf(buf + len, "%02hhX", *command++);
 	len += sprintf(buf + len, "\"");
 
-	ret = g_at_chat_send(sd->chat, buf, none_prefix,
-				mbm_stkr_cb, cbd, g_free);
+	DBG("%s", buf);
 
-	g_free(buf);
-	buf = NULL;
-
-	if (ret > 0)
+	if (g_at_chat_send(sd->chat, buf, none_prefix,
+					stkr_cb, cbd, g_free) > 0) {
+		g_free(buf);
 		return;
+	}
 
 error:
 	g_free(buf);
@@ -158,6 +164,8 @@ static void stki_notify(GAtResult *result, gpointer user_data)
 	const guint8 *pdu;
 	gint len;
 
+	DBG("");
+
 	g_at_result_iter_init(&iter, result);
 
 	if (!g_at_result_iter_next(&iter, "*STKI:"))
@@ -171,16 +179,34 @@ static void stki_notify(GAtResult *result, gpointer user_data)
 
 static void stkn_notify(GAtResult *result, gpointer user_data)
 {
-	/* Proactive command has been handled by the modem.  Should
-	 * the core be notified?  For now we just ignore it because
-	 * we must not respond to the command.
-	 */
+	struct ofono_stk *stk = user_data;
+	GAtResultIter iter;
+	const guint8 *pdu;
+	gint len;
+
+	DBG("");
+
+	/* Proactive command has been handled by the modem. */
+	g_at_result_iter_init(&iter, result);
+
+	if (g_at_result_iter_next(&iter, "*STKN:") == FALSE)
+		return;
+
+	if (g_at_result_iter_next_hexstring(&iter, &pdu, &len) == FALSE)
+		return;
+
+	if (len == 0)
+		return;
+
+	ofono_stk_proactive_command_handled_notify(stk, len, pdu);
 }
 
 static void stkend_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_stk *stk = user_data;
 
+	DBG("");
+
 	ofono_stk_proactive_session_end_notify(stk);
 }
 
@@ -189,6 +215,8 @@ static void mbm_stkc_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct ofono_stk *stk = user_data;
 	struct stk_data *sd = ofono_stk_get_data(stk);
 
+	DBG("");
+
 	if (!ok)
 		return;
 
@@ -205,8 +233,13 @@ static int mbm_stk_probe(struct ofono_stk *stk, unsigned int vendor, void *data)
 	GAtChat *chat = data;
 	struct stk_data *sd;
 
-	sd = g_new0(struct stk_data, 1);
-	sd->chat = chat;
+	DBG("");
+
+	sd = g_try_new0(struct stk_data, 1);
+	if (!sd)
+		return -ENOMEM;
+
+	sd->chat = g_at_chat_clone(chat);
 
 	ofono_stk_set_data(stk, sd);
 
@@ -221,8 +254,11 @@ static void mbm_stk_remove(struct ofono_stk *stk)
 {
 	struct stk_data *sd = ofono_stk_get_data(stk);
 
+	DBG("");
+
 	ofono_stk_set_data(stk, NULL);
 
+	g_at_chat_unref(sd->chat);
 	g_free(sd);
 }
 
